''*******************************************************
''*  TV Driver with minimal hub memory use  v2.0        *
''*  Author: Chip Gracey                                *
''*  Modifications by: Michael Green                    *
''*  Copyright (c) 2004 Parallax, Inc.                  *
''*  Modifications copyright (c) 2010 by Michael Green  *
''*  See end of file for terms of use.                  *               
''*******************************************************

' v1.0 - 01 May 2006 - original version
' v1.1 - 17 May 2006 - pixel tile size can now be 16 x 32 to enable more
'                      efficient character displays utilizing the internal
'                      font.  See 'tv_mode'
' v2.0 - 12 Jan 2010 - rewritten based on Michael Park's driver that places
'                      the display buffer in the cog's memory.  The user
'                      interface has changed to allow increased functionality.
'                      The current display row (1-13) is in interface.byte[3].
'                      The current display column (1-40) is in interface.byte[2]
'                      The interface.byte[1] is a command code and
'                      interface.byte[0] is an operand for the command.
'                      The low order 4 or 6 bits of the command may be a repeat
'                      count.  If repeat count is zero, nothing is done
'                      Commands: $00     - Idle
'                                $01     - Copy display char to operand, C++
'                                $02     - Disable display driver I/O pins
'                                $03     - Enable display driver I/O pins
'                                $04     - Set foreground color from operand
'                                $05     - Set background color from operand
'                                $06-$0F - Reserved (ignored)
'                                $10-$1F - Fill current row with operand, R++
'                                $20-$2F - Move next row upwards, R++
'                                $30-$3F - Move previous row downwards, R--
'                                $40-$7F - Output operand to display, C++
'                                $80-$BF - Move next character to left, C++
'                                $C0-$FF - Move previous character to right, C--
'                      Note: The current display column and row are updated
'                            after each occurance of an operation as noted.
'                            The command byte is zeroed when completed.
'                      Note: This is expected to provide a 40 x 13 display.
'                            There's an additional blank line above and below
'                            the display area to facilitate scrolling.  Row 0 is
'                            the blank row above the display area and row 14 is
'                            the blank row below the display area.  Column 0 is
'                            the last column of the previous row and column 41
'                            is the first column of the next row.

OBJ
  dsp : "Text_TV"                 'programming interface for driver

CON
  fntsc         = 3_579_545       'NTSC color frequency
  lntsc         = 3640            'NTSC color cycles per line * 16
  sntsc         = 624             'NTSC color cycles per sync * 16

  fpal          = 4_433_618       'PAL color frequency
  lpal          = 4540            'PAL color cycles per line * 16
  spal          = 848             'PAL color cycles per sync * 16

  cols          = dsp#cols        'number of columns on a display line
  colsExpand    = 4               ' expansion factor
  colsOffset    = 0               ' column offset
  rows          = dsp#rows        'number of display lines on the display
  rowsExpand    = 1               ' expansion factor
  rowsOffset    = 0               ' row offset

  Bsp           = dsp#BS          'backspace
  Tab           = dsp#HT          'horizontal tab
  Lf            = dsp#LF          'line feed
  Cr            = dsp#CR          'carriage return
  Esc           = dsp#ESC         'escape

  doNothing     = dsp#doNothing   'do nothing
  getDispChar   = dsp#getDispChar 'copy display character to operand, C++
  disableDrvr   = dsp#disableDrvr 'disable display driver I/O pins
  enableDrvr    = dsp#enableDrvr  'enable display driver I/O pins
  setForeClr    = dsp#setForeClr  'set foreground color from operand
  setBackClr    = dsp#setBackClr  'set background color from operand
  fillRow       = dsp#fillRow     'fill current row with operand, R++
  moveRowUp     = dsp#moveRowUp   'move next row upwards, R++
  moveRowDown   = dsp#moveRowDown 'move previous row downwards, R--
  putDispChar   = dsp#putDispChar 'output operand to display, C++
  deleteLeft    = dsp#deleteLeft  'move next character to left, C++
  deleteRight   = dsp#deleteRight 'move previous character to right, C--
  
VAR
  long  cog

PUB start(pinInfo,modeInfo,aural,freq,outBlk,ctrlBlk) 'initialize the driver
' pins    = video pins information (see tv_pins description below)
' mode    = video mode information (see tv_mode description below)
' aural   = cog number (0-7) used for broadcast audio (see tv_auralcog)
' freq    = channel # used for broadcast frequency (see tv_broadcast)
'           0 - none, 1 to 5 - broadcast channels 2 to 6
' outBlk  = address of long used to communicate with VT100 interpreter
' ctrlBlk = address of long used to communicate with I/O driver
   stop
   long[ctrlBlk] := freq<<24 | aural<<16 | modeInfo<<8 | pinInfo
   result := cog := cognew(@cogEntry,ctrlBlk) + 1
   dsp.start(outBlk,ctrlBlk)                            'initialize interface

PUB stop                                                'stop the TV driver
   if cog
      cogstop(cog~ - 1)

PUB str(s)                                              'display string at s
   dsp.str(s)

PUB dec(v)                                              'display decimal value
   dsp.dec(v)
   
PUB hex(h,p)                                            'display hexadecimal
   dsp.hex(h,p)                                         ' value, width w

PUB bin(h,p)                                            'display binary value
   dsp.bin(h,p)                                         ' width w
   
PUB out(c)                                              'display single char
   dsp.out(c)

PUB blink(t)                                            'blink cursor, t in ms
   dsp.blink(t)                                         ' t=-1 start, t=0 stop
   
DAT
                        org     0
' Cog entry

cogEntry                call    #text                   'do once-only init
                        movd    :fillBlks,#text
                        mov     temp,#(rows+2)*cols/4   'fill buffer with blanks
:fillBlks               mov     0-0,blanks
                        add     :fillBlks,d0
                        djnz    temp,#:fillBlks
                        mov     taskptr,#endOfCmd       'vert retrace tasks

' Superfield processing

superfield              test    mode,#%00001    wc      'if ntsc, set phaseflip
        if_nc           mov     phaseflip,phasemask
                        test    mode,#%00010    wc      'get interlace into c
                        muxnc   interlace,#$FF  wz,nr   ' then copy !c to z

' Field processing

field                   mov     x,vinv                  'inv. back porch lines
:black                  call    #hsync                  'do hsync
                        waitvid burst,sync_high2        'do black
                        jmpret  taskret,taskptr         'call task section
                        djnz    x,#:black               'another black line?
                        mov     x,vb                    'do visible back
                        call    #blank_lines            ' porch lines
                        mov     y,#rows                 'set vertical tiles
                        movs    :tile,#text + cols/4    'point to first char
:line                   mov     w,#32                   'set lines per char
                        mov     fontTbl,fontTbl0        'start of font table
:vert   if_z            xor     interlace,#1            'interlace skip?
        if_z            tjz     interlace,#:skip
                        call    #hsync                  'do hsync
                        mov     vscl,hb                 'do visible back
                        xor     tile,colortable         ' porch pixels
                        waitvid tile,#0
                        mov     x,#cols/4               'set horizontal tiles/4
                        mov     vscl,hx                 'set horizontal expand
:tile                   mov     chars4,0-0              'get character's long
                        mov     z,#4                    ' and position in long
:loop4                  mov     charTemp,chars4         'get next character
                        ror     chars4,#8
                        and     charTemp,#$FF
                        shr     charTemp,#1     wc      'odd/even char code?
                        shl     charTemp,#7             'position for lookup
                        add     charTemp,fontTbl        ' in font table
                        rdlong  pixels16,charTemp       'get scan line for char
        if_nc           waitvid colortable+0,pixels16
        if_c            waitvid colortable+1,pixels16   'adjust for odd/even
                        djnz    z,#:loop4               'continue with chars
                        add     :tile,#1                'continue with longs
                        djnz    x,#:tile
                        mov     vscl,hf                 'do visible front
                        mov     tile,phaseflip          ' porch pixels
                        xor     tile,colortable+0
                        waitvid tile,#0
                        sub     :tile,#cols/4           'repeat as needed for
:skip                   add     fontTbl,#4              ' remaining font lines
                        djnz    w,#:vert
                        add     :tile,#cols/4
                        djnz    y,#:line                'another tile line?
        if_z            xor     interlace,#1    wz      'set interlace & field1
                        test    mode,#%00001    wc      'do visible front
                        mov     x,vf                    ' porch lines
        if_nz_and_c     add     x,#1
                        call    #blank_lines            'invisible part of scan
        if_z_eq_c       call    #hsync
        if_z_eq_c       mov     vscl,hrest              'do short line if needed
        if_z_eq_c       waitvid burst,sync_high2
        if_z_eq_c       xor     phaseflip,phasemask
                        call    #vsync_high             'do high vsync pulses
                        movs    vsync1,#sync_low1       'do low vsync pulses
                        movs    vsync2,#sync_low2
                        call    #vsync_low
                        call    #vsync_high             'do high vsync pulses
        if_nz           mov     vscl,hhalf              'odd frame? do half line
        if_nz           waitvid burst,sync_high2
        if_z            jmp     #field                  'if interlace and field1
                        jmp     #superfield             ' field2 else superfield

' Blank lines

blank_lines             call    #hsync                  'do hsync
                        xor     tile,colortable         'do background
                        waitvid tile,#0
                        djnz    x,#blank_lines
blank_lines_ret         ret

' Horizontal sync

hsync                   test    mode,#%00001    wc      'pal? toggle phaseflip
        if_c            xor     phaseflip,phasemask
                        mov     vscl,sync_scale1        'do hsync       
                        mov     tile,phaseflip
                        xor     tile,burst
                        waitvid tile,sync_normal
                        mov     vscl,hvis               'setup in case blank line
                        mov     tile,phaseflip
hsync_ret               ret

' Vertical sync

vsync_high              movs    vsync1,#sync_high1      'vertical sync
                        movs    vsync2,#sync_high2
vsync_low               mov     x,vrep
vsyncx                  mov     vscl,sync_scale1
vsync1                  waitvid burst,sync_high1
                        mov     vscl,sync_scale2
vsync2                  waitvid burst,sync_high2
                        djnz    x,#vsyncx
vsync_low_ret
vsync_high_ret          ret

' Vertical blanking interval tasks
'
waitForCmd              djnz    count,#waitPause        'see if count goes zero
endOfCmd                mov     temp,row                'pass back row & col
                        shl     temp,#8
                        add     temp,col
                        shl     temp,#16
                        add     temp,char               'pass back any result
                        wrlong  temp,par
fetchNewCmd             jmpret  taskptr,taskret
                        rdlong  char,par                'get request to do
                        mov     count,char
                        and     char,#$FF
                        shr     count,#8                'break it into pieces
                        mov     command,count
                        and     command,#$FF
                        mov     col,count
                        and     count,#$3F
                        shr     col,#8
                        mov     row,col
                        and     col,#$3F
                        shr     row,#8
                        and     row,#$0F
                        cmp     command,#putDispChar wc 'adjust count for
        if_c            and     count,#$0F              ' command size
                        cmp     command,#fillRow wc     'force count to 1
        if_c            mov     count,#$01
                        tjz     command,#fetchNewCmd    'idle command
                        tjz     count,#endOfCmd         'zero count in command
                        jmp     #waitContinue

waitPause               jmpret  taskptr,taskret
waitContinue            mov     offset,row
                        shl     offset,#2               'multiply row by 10 to
                        add     offset,row              ' get offset in longs
                        shl     offset,#1
                        add     offset,#text            'add address of text
                        mov     temp,col                'adjust address for col
                        sub     temp,#1
                        sar     temp,#2
                        add     offset,temp
                        mov     shift,col               'get shift count for col
                        sub     shift,#1
                        and     shift,#%11
                        shl     shift,#3
                        mov     charMask,#$FF           'make character mask
                        shl     charMask,shift
                        cmp     command,#deleteRight wc
        if_nc           jmp     #moveRight              'move line to right
                        cmp     command,#deleteLeft wc
        if_nc           jmp     #moveLeft               'move line to left
                        cmp     command,#putDispChar wc
        if_nc           jmp     #dispChar               'display character
                        cmp     command,#moveRowDown wc
        if_nc           jmp     #moveDown               'move lines down
                        cmp     command,#moveRowUp wc
        if_nc           jmp     #moveUp                 'move lines up
                        cmp     command,#fillRow wc
        if_nc           jmp     #fillLine               'fill line with char
                        cmp     command,#setBackClr+1 wc
        if_nc           jmp     #waitForCmd             'unused commands $06-$0F
                        cmp     command,#setBackClr wc
        if_nc           jmp     #setBackgrnd            'set background color
                        cmp     command,#setForeClr wc
        if_nc           jmp     #setForegrnd            'set foreground color
                        cmp     command,#enableDrvr wc
        if_nc           jmp     #enableCmd              'enable driver I/O pins
                        cmp     command,#disableDrvr wc
        if_nc           jmp     #disableCmd             'disable driver I/O pins
                        cmp     command,#getDispChar wc
        if_nc           jmp     #copyChar               'copy display to operand
                        jmp     #waitForCmd

dispChar                movd    :writeChar1,offset      'address of current char
                        movd    :writeChar2,offset
                        mov     temp,char
                        shl     temp,shift              'position character
:writeChar1             andn    0-0,charMask
:writeChar2             or      0-0,temp                'insert char in buffer
                        add     col,#1                  'increment column
                        jmp     #waitForCmd

copyChar                movs    :readChar1,offset       'address of current char
                        nop
:readChar1              mov     char,0-0                'get char from buffer,
                        and     char,charMask           ' isolate it, and
                        shr     char,shift              '  return as status
                        add     col,#1                  'increment column
                        jmp     #waitForCmd

moveRight               cmp     shift,#8        wc      'get previous character
        if_c            sub     offset,#1               'may need previous long
                        movs    :readChar1,offset
                        ror     charMask,#8
:readChar1              mov     temp,0-0
                        and     temp,charMask
                        rol     temp,#8                 'move character to right
        if_c            add     offset,#1
                        rol     charMask,#8
                        movd    :writeChar1,offset
                        movd    :writeChar2,offset
:writeChar1             andn    0-0,charMask
:writeChar2             or      0-0,temp                'insert char in buffer
                        sub     col,#1                  'decrement column
                        jmp     #waitForCmd

moveLeft                cmp     shift,#24       wc      'get next character
        if_nc           add     offset,#1               'may need next long
                        movs    :readChar1,offset
                        rol     charMask,#8
:readChar1              mov     temp,0-0
                        and     temp,charMask
                        ror     temp,#8                 'move character to left
        if_nc           sub     offset,#1
                        ror     charMask,#8
                        movd    :writeChar1,offset
                        movd    :writeChar2,offset
:writeChar1             andn    0-0,charMask
:writeChar2             or      0-0,temp                'insert char in buffer
                        add     col,#1                  'increment column
                        jmp     #waitForCmd

moveDown                movd    :move4chars,offset      'copy one row from the
                        sub     offset,#cols/4          ' previous row to the
                        movs    :move4chars,offset      '  current one
                        mov     temp,#cols/4
:move4chars             mov     0-0,0-0
                        add     :move4chars,d0s0
                        djnz    temp,#:move4chars
                        sub     row,#1                  'decrement row
                        jmp     #waitForCmd

moveUp                  movd    :move4chars,offset      'copy one row from the
                        add     offset,#cols/4          ' next row to the
                        movs    :move4chars,offset      '  current one
                        mov     temp,#cols/4
:move4chars             mov     0-0,0-0
                        add     :move4chars,d0s0
                        djnz    temp,#:move4chars
                        add     row,#1                  'increment row
                        jmp     #waitForCmd

fillLine                mov     fillChar,char           'use char for fill
                        shl     fillChar,#8
                        or      fillChar,char           'duplicate to fill long
                        shl     fillChar,#8
                        or      fillChar,char
                        shl     fillChar,#8
                        or      fillChar,char
                        movd    :fill4chars,offset
                        mov     temp,#cols/4
:fill4chars             mov     0-0,fillChar            'copy to display line
                        add     :fill4chars,d0
                        djnz    temp,#:fill4chars
                        add     row,#1
                        jmp     #waitForCmd

setForegrnd             mov     foreground,char
                        jmp     #setColors

setBackgrnd             mov     background,char
setColors               mov     temp,foreground
                        shl     temp,#8
                        add     temp,background
                        mov     colortable+0,temp
                        shl     colortable+0,#16
                        add     colortable+0,temp
                        mov     temp,foreground
                        shl     temp,#16
                        add     temp,background
                        mov     colortable+1,temp
                        shl     colortable+1,#8
                        add     colortable+1,temp
                        jmp     #waitForCmd


enableCmd               mov     dira,saveDira           're-enable driver
                        mov     dirb,saveDirb
                        jmp     #waitForCmd

disableCmd              mov     dira,#0                 'disable driver
                        mov     dirb,#0
                        jmp     #waitForCmd

' Initialized data
'
count                   long    1
m8                      long    8_000_000
m128                    long    128_000_000
d0                      long    1 << 9 << 0
d6                      long    1 << 9 << 6
d0s0                    long    1 << 9 << 0 + 1 << 0
d0s1                    long    1 << 9 << 0 + 1 << 1
phaseflip               long    $00000000
phasemask               long    $F0F0F0F0
sync_high1              long    %0101010101010101010101_101010_0101
sync_high2              long    %01010101010101010101010101010101 'used for blk
sync_low1               long    %1010101010101010101010101010_0101
sync_low2               long    %01_101010101010101010101010101010
col                     long    1                       'current display col
row                     long    rows                    'current display row
foreground              long    $05                     'white foreground
background              long    $02                     'black background
fontTbl0                long    $8000                   'start of font table
colortable              long    $05_02_05_02            'white on black bkgnd
                        long    $05_05_02_02

' Uninitialized data
'
command                 long    0
offset                  long    0                       'char offset into buffer
shift                   long    0                       'shift count for col
charMask                long    0                       'for masking characters
fillChar                long    0
char                    long    0
interlace               long    0

taskptr                 long    0                       'tasks
taskret                 long    0
temp                    long    0

w                       long    0
x                       long    0                       'display
y                       long    0
z                       long    0
hf                      long    0
hb                      long    0
hx                      long    0
vf                      long    0
vb                      long    0
vx                      long    0
tile                    long    0
charTemp                long    0
fontTbl                 long    0
chars4                  long    0
pixels16                long    0
mode                    long    0                       'video mode (%tccip)

hvis                    long    0                       'loaded from word table
hrest                   long    0
hhalf                   long    0
vvis                    long    0
vinv                    long    0
vrep                    long    0
burst                   long    0

fcolor                  long    0                       'loaded from long table
sync_scale1             long    0
sync_scale2             long    0
sync_normal             long    0

saveDira                long    0
saveDirb                long    0
                        
blanks                  long    $20_20_20_20            'must come just before text

' text buffer (used for once-only initialization)
'
text                    rdlong  t1,par                  'get initial parameters
                        mov     pins,t1
                        and     pins,#$FF               'LSB is video pins
                        shr     t1,#$8
                        mov     mode,t1
                        and     mode,#$FF               'next byte is mode
                        shr     t1,#$8
                        mov     auralcog,t1
                        and     auralcog,#$FF           'next byte is auralcog
                        shr     t1,#$8
                        add     t1,#freqTable
                        movs    :moveFreq,t1            'MSB is index into
                        mov     vx,#rowsExpand          ' frequency table
:moveFreq               mov     broadcast,0-0
                        mov     t1,pins                 'set video pins & dirs
                        test    t1,#$08         wc
        if_nc           mov     t2,pins0
        if_c            mov     t2,pins1
                        test    t1,#$40         wc
                        shr     t1,#1
                        shl     t1,#3
                        shr     t2,t1
                        movs    vcfg,t2
                        shr     t1,#6
                        movd    vcfg,t1
                        shl     t1,#3
                        and     t2,#$FF
                        shl     t2,t1
        if_nc           mov     saveDira, t2
        if_nc           mov     saveDirb, #0
        if_c            mov     saveDira, #0
        if_c            mov     saveDirb, t2
                        movs    :rdTbl,#wtab            'load ntsc/pal metrics
                        movd    :wrTbl,#hvis            ' from word table
                        mov     t1,#wtabx - wtab
                        test    mode,#%00001    wc
:rdTbl                  mov     t2,0-0
                        add     :rdTbl,#1
        if_nc           shl     t2,#16
                        shr     t2,#16
:wrTbl                  mov     0-0,t2
                        add     :wrTbl,d0
                        djnz    t1,#:rdTbl              '+54
        if_nc           movs    :ltab,#ltab             'load ntsc/pal metrics
        if_c            movs    :ltab,#ltab+1           ' from long table
                        movd    :ltab,#fcolor
                        mov     t1,#(ltabx - ltab) >> 1
:ltab                   mov     0-0,0-0
                        add     :ltab,d0s1
                        djnz    t1,#:ltab               '+17
                        rdlong  t1,#0                   'get CLKFREQ
                        shr     t1,#1                   'if CLKFREQ < 16MHz,
                        cmp     t1,m8           wc      ' cancel broadcast
        if_c            mov     broadcast,#0
                        shr     t1,#1                   'if CLKFREQ < color
                        cmp     t1,fcolor       wc      ' freq * 4, disable
        if_c            jmp     #text_ret               '+11
                        mov     t1,fcolor               'ctra pll = fcolor * 16
                        call    #divide                 'ntsc: vco = fcolor * 32
                        test    mode,#%00001    wc      'pal: vco = fcolor * 16
        if_c            movi    ctra,#%00001_111        'select fcolor * 16 out
        if_nc           movi    ctra,#%00001_110        ' (ntsc=/2, pal=/1)
        if_nc           shl     t2,#1
                        mov     frqa,t2                 '+147
                        mov     t1,broadcast            'ctrb pll to _broadcast
                        mov     t2,#0                   'if 0, turn off ctrb
                        tjz     t1,#:off
                        min     t1,m8                   'limit 8MHz to 128MHz
                        max     t1,m128
                        mov     t2,#%00001_100          '_broadcast -> 4MHz-8MHz
:scale                  shr     t1,#1                   ' (vco -> 64MHz-128MHz)
                        cmp     m8,t1           wc
        if_c            add     t2,#%00000_001
        if_c            jmp     #:scale
:off                    movi    ctrb,t2
                        call    #divide
                        mov     frqb,t2                 '+165
                        mov     t1,#%10100_000          'set video configuration
                        test    pins,#$01       wc      '(broadcast/baseband
        if_c            or      t1,#%01000_000          '  swap output bits?)
                        test    mode,#%01000    wc      '(strip chroma from
        if_nc           or      t1,#%00010_000          '  broadcast?)
                        test    mode,#%00100    wc      '(strip chroma from
        if_nc           or      t1,#%00001_000          '  baseband?)
                        and     auralcog,#%111          '(set aural cog)
                        or      t1,auralcog
                        movi    vcfg,t1                 '+10
                        mov     hx,#colsExpand          'compute horiz metrics
                        shl     hx,#8
                        or      hx,#colsExpand
                        shl     hx,#4
                        mov     t1,#cols
                        mov     t2,#colsExpand
                        call    #multiply
                        mov     hf,hvis
                        sub     hf,t1
                        shr     hf,#1           wc
                        mov     hb,#colsOffset
                        addx    hb,hf
                        sub     hf,#colsOffset          '+52
                        mov     t1,#rows                'compute vert metrics
                        mov     t2,#rowsExpand
                        call    #multiply
                        test    mode,#%10000    wc      'consider tile size
        if_c            shl     t1,#1
                        test    mode,#%00010    wc      'consider interlace
        if_c            shr     t1,#1
                        mov     vf,vvis
                        sub     vf,t1
                        shr     vf,#1           wc
                        neg     vb,#rowsOffset
                        addx    vb,vf
                        add     vf,#rowsOffset          '+53
                        mov     dira,saveDira           'enable I/O pins
                        mov     dirb,saveDirb
text_ret                ret

' Divide t1/CLKFREQ to get frqa or frqb value into t2
'
divide                  rdlong  m1,#0                   'get CLKFREQ
                        mov     m2,#32+1
:loop                   cmpsub  t1,m1           wc
                        rcl     t2,#1
                        shl     t1,#1
                        djnz    m2,#:loop
divide_ret              ret                             '+140

' Multiply t1 * t2 * 16 (t1, t2 = bytes)
'
multiply                shl     t2,#8+4-1
                        mov     m1,#8
:loop                   shr     t1,#1           wc
        if_c            add     t1,t2
                        djnz    m1,#:loop
multiply_ret            ret                             '+37

pins                    long    0       '%pppmmmm       'starting I/O pin
broadcast               long    0       '0+             'broadcast mode
auralcog                long    0       '0-7            'aural cog

' NTSC/PAL metrics tables
'                               ntsc                    pal
'                              ----------------------------------------------
wtab                    word    lntsc - sntsc,          lpal - spal     'hvis
                        word    lntsc / 2 - sntsc,      lpal / 2 - spal 'hrest
                        word    lntsc / 2,              lpal / 2        'hhalf
                        word    243,                    286             'vvis
                        word    10,                     18              'vinv
                        word    6,                      5               'vrep
                        word    $02_8A,                 $02_AA          'burst
wtabx
ltab                    long    fntsc                                   'fcolor
                        long    fpal
                        long    sntsc >> 4 << 12 + sntsc           'sync_scale1
                        long    spal >> 4 << 12 + spal
                        long    67 << 12 + lntsc / 2 - sntsc       'sync_scale2
                        long    79 << 12 + lpal / 2 - spal
                        long    %0101_00000000_01_10101010101010_0101  'normal
                        long    %010101_00000000_01_101010101010_0101
ltabx

pins0                   long    %11110000_01110000_00001111_00000111
pins1                   long    %11111111_11110111_01111111_01110111

freqTable               long    0                       'broadcast not used
                        long    55_250_000              'channel 2
                        long    61_250_000              'channel 3
                        long    67_250_000              'channel 4
                        long    77_250_000              'channel 5
                        long    83_250_000              'channel 6

t1                      long    0                       'used during init only
t2                      long    0
t3                      long    0
m1                      long    0
m2                      long    0

''Parameter descriptions:
''  _________
''  tv_pins
''
''    bits 6..4 select pin group:
''      %000: pins 7..0
''      %001: pins 15..8
''      %010: pins 23..16
''      %011: pins 31..24
''      %100: pins 39..32
''      %101: pins 47..40
''      %110: pins 55..48
''      %111: pins 63..56
''
''    bits 3..0 select pin group mode:
''      %0000: %0000_0111    -                    baseband
''      %0001: %0000_0111    -                    broadcast
''      %0010: %0000_1111    -                    baseband + chroma
''      %0011: %0000_1111    -                    broadcast + aural
''      %0100: %0111_0000    broadcast            -
''      %0101: %0111_0000    baseband             -
''      %0110: %1111_0000    broadcast + aural    -
''      %0111: %1111_0000    baseband + chroma    -
''      %1000: %0111_0111    broadcast            baseband
''      %1001: %0111_0111    baseband             broadcast
''      %1010: %0111_1111    broadcast            baseband + chroma
''      %1011: %0111_1111    baseband             broadcast + aural
''      %1100: %1111_0111    broadcast + aural    baseband
''      %1101: %1111_0111    baseband + chroma    broadcast
''      %1110: %1111_1111    broadcast + aural    baseband + chroma
''      %1111: %1111_1111    baseband + chroma    broadcast + aural
''      -----------------------------------------------------------
''            active pins    top nibble           bottom nibble
''
''      the baseband signal nibble is arranged as:
''        bit 3: chroma signal for s-video (attach via 560-ohm resistor)
''        bits 2..0: baseband video (sum 270/560/1100-ohm resistors to form
''                   75-ohm 1V signal)
''
''      the broadcast signal nibble is arranged as:
''        bit 3: aural subcarrier (sum 560-ohm resistor into network below)
''        bits 2..0: visual carrier (sum 270/560/1100-ohm resistors to form
''                   75-ohm 1V signal)
''  _______
''  tv_mode
''
''    bit 4 selects between 16x16 and 16x32 pixel tiles:
''      0: 16x16 pixel tiles (tileheight = 16)
''      1: 16x32 pixel tiles (tileheight = 32)
''
''    bit 3 controls chroma mixing into broadcast:
''      0: mix chroma into broadcast (color)
''      1: strip chroma from broadcast (black/white)
''
''    bit 2 controls chroma mixing into baseband:
''      0: mix chroma into baseband (composite color)
''      1: strip chroma from baseband (black/white or s-video)
''
''    bit 1 controls interlace:
''      0: progressive scan (243 display lines for NTSC, 286 for PAL)
''           less flicker, good for motion
''      1: interlaced scan (486 display lines for NTSC, 572 for PAL)
''           doubles the vertical display lines, good for text
''
''    bit 0 selects NTSC or PAL format
''      0: NTSC
''           3016 horizontal display ticks
''           243 or 486 (interlaced) vertical display lines
''           CLKFREQ must be at least 14_318_180 (4 * 3_579_545 Hz)*
''      1: PAL
''           3692 horizontal display ticks
''           286 or 572 (interlaced) vertical display lines
''           CLKFREQ must be at least 17_734_472 (4 * 4_433_618 Hz)*
''
''      * driver will disable itself while CLKFREQ is below requirement
''  ______________________________
''  foreground / background colors
''
''      8-bit color fields are as follows:
''        bits 7..4: chroma data (0..15 = blue..green..red..)*
''        bit 3: controls chroma modulation (0=off, 1=on)
''        bits 2..0: 3-bit luminance level:
''          values 0..1: reserved for sync - don't use
''          values 2..7: valid luminance range, modulation adds/subtracts 1
'                        (beware of 7)
''                       value 0 may be modulated to produce a saturated color
'                        toggling between levels 1 and 7
''
''      * because of TV's limitations, it doesn't look good when chroma changes abruptly -
''        rather, use luminance - change chroma only against a black or white background for
''        best appearance
''  _____
''  tv_ht
''
''    horizontal number pixel tiles - must be at least 1
''    practical limit is 40 for NTSC, 50 for PAL
''  _____
''  tv_vt
''
''    vertical number of pixel tiles - must be at least 1
''    practical limit is 13 for NTSC, 15 for PAL (26/30 max for interlaced
''    NTSC/PAL)
''  _____
''  tv_hx
''
''    horizontal tile expansion factor - must be at least 3 for NTSC, 4 for PAL
''
''    make sure 16 * tv_ht * tv_hx + ||tv_ho + 32 is less than the horizontal
''    display ticks
''  _____
''  tv_vx
''
''    vertical tile expansion factor - must be at least 1
''
''    make sure <tileheight> * tv_vt * tv_vx + ||tv_vo + 1 is less than the
''    display lines
''  _____
''  tv_ho
''
''    horizontal offset in ticks - pos/neg value (0 for centered image)
''    shifts the display right/left
''  _____
''  tv_vo
''
''    vertical offset in lines - pos/neg value (0 for centered image)
''    shifts the display up/down
''  ____________
''  tv_broadcast
''
''    broadcast frequency expressed in Hz (ie channel 2 is 55_250_000)
''    if 0, modulator is turned off - saves power
''
''    broadcasting requires CLKFREQ to be at least 16_000_000
''    while CLKFREQ is below 16_000_000, modulator will be turned off
''  ___________
''  tv_auralcog
''
''    selects cog to supply aural fm signal - 0..7
''    uses ctra pll output from selected cog
''
''    in NTSC, the offset frequency must be 4.5MHz and the max bandwidth +-25KHz
''    in PAL, the offset frequency and max bandwidth vary by PAL type

{{
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}}
