''*******************************************************
''*  TV Driver with minimal hub memory use  v2.0        *
''*  Author: Chip Gracey                                *
''*  Modifications by: Michael Green                    *
''*  Copyright (c) 2004 Parallax, Inc.                  *
''*  Modifications copyright (c) 2010 by Michael Green  *
''*  See end of file for terms of use.                  *               
''*******************************************************

' v1.0 - 01 May 2006 - original version
' v1.1 - 17 May 2006 - pixel tile size can now be 16 x 32 to enable more
'                      efficient character displays utilizing the internal
'                      font.  See 'tv_mode'
' v2.0 - 12 Jan 2010 - rewritten based on Michael Park's driver that places
'                      the display buffer in the cog's memory.  The user
'                      interface has changed to allow increased functionality.
'                      The current display row (1-13) is in interface.byte[3].
'                      The current display column (1-40) is in interface.byte[2]
'                      The interface.byte[1] is a command code and
'                      interface.byte[0] is an operand for the command.
'                      The low order 4 or 6 bits of the command may be a repeat
'                      count.  If repeat count is zero, nothing is done
'                      Commands: $00     - Idle
'                                $01     - Copy display char to operand, C++
'                                $02     - Disable display driver I/O pins
'                                $03     - Enable display driver I/O pins
'                                $04     - Set foreground color from operand
'                                $05     - Set background color from operand
'                                $06-$0F - Reserved (ignored)
'                                $10-$1F - Fill current row with operand, R++
'                                $20-$2F - Move next row upwards, R++
'                                $30-$3F - Move previous row downwards, R--
'                                $40-$7F - Output operand to display, C++
'                                $80-$BF - Move next character to left, C++
'                                $C0-$FF - Move previous character to right, C--
'                      Note: The current display column and row are updated
'                            after each occurance of an operation as noted.
'                            The command byte is zeroed when completed.
'                      Note: This is expected to provide a 40 x 13 display.
'                            There's an additional blank line above and below
'                            the display area to facilitate scrolling.  Row 0 is
'                            the blank row above the display area and row 14 is
'                            the blank row below the display area.  Column 0 is
'                            the last column of the previous row and column 41
'                            is the first column of the next row.

CON
  cols          = 40            'number of columns on a display line
  rows          = 13            'number of display lines on the display

  #$00                          'ANSI Control Codes
  NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS,  HT,  LF,  VT,  FF,  CR,  SO,  SI
  DLE, DC1, DC2, DC3, DC4, NAK, SYN, ETB, CAN, EM, xSUB, ESC, FS,  GS,  RS,  US
  #$7F
  DEL

  Bsp           = BS
  Tab           = HT
  
  doNothing     = $00                  'do nothing
  getDispChar   = $01                  'copy display character to operand, C++
  disableDrvr   = $02                  'disable display driver I/O pins
  enableDrvr    = $03                  'enable display driver I/O pins
  setForeClr    = $04                  'set foreground color from operand
  setBackClr    = $05                  'set background color from operand
  fillRow       = $10                  'fill current row with operand, R++
  moveRowUp     = $20                  'move next row upwards, R++
  moveRowDown   = $30                  'move previous row downwards, R--
  putDispChar   = $40                  'output operand to display, C++
  deleteLeft    = $80                  'move next character to left, C++
  deleteRight   = $C0                  'move previous character to right, C--

  maxVt100      = 4                    'maximum number of VT100 parameters
  cursorChar    = $5F                  'cursor character

VAR
   long save, time, vt100Val[maxVt100]
   byte state, saved, vt100Cnt, vt100Def[maxVt100]

PUB start(outBlk,ctrlBlk)
   save := ctrlBlk                     'save address of control long
   state~                              'main state
   saved~                              'no saved character

PUB blink(t) | ms                      'blink a cursor, t in ms
   if t == false                       ' if t == 0, restore char
      if saved                         '  restore any character
         waitForCog
         word[save] := (putDispChar+1)<<8 + saved~
         waitForCog
         byte[save+2]--                '  restore the position
      return
   if t == true                        ' if t == -1, set initial time
      time := cnt
      return
   t *= clkfreq / 1000
   repeat while cnt-time > t           ' while elapsed time > t
      time += t                        '  bring saved time up to date
      waitForCog
      if saved                         '  restore any character
         word[save] := (putDispChar+1)<<8 + saved~
      else                             '  blink at cursor
         word[save] := constant(getDispChar<<8)
         waitForCog                    '  save current character
         saved := byte[save]
         byte[save+2]--                '  overstore with cursor
         word[save] := constant((putDispChar+1)<<8 + cursorChar)
      waitForCog
      byte[save+2]--                   '  restore the position
   
PUB str(s)                             'display string at s
   repeat strsize(s)
      out(byte[s++])

PUB dec(v) | i, x                      'display decimal number
   x := v == NEGX                      'check for max negative
   if v < 0
      v := ||(v+x)                     'adjust as needed
      out("-")                         ' and output sign
   i := 1_000_000_000                  'initialize divisor
   repeat 10                           'loop for 10 digits
      if v => i                                                               
         out(v / i + "0" + x*(i == 1)) 'display digit
         v //= i
         result~~                      'flag non-zero found
      elseif result or i == 1
         out("0")                      'display at least zero
      i /= 10

PUB hex(v,w)                           'display hexadecimal
   v <<= (8 - w) << 2                  ' at most 8 digits
   repeat w
      out(lookupz((v <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(v, w)                          'display binary
   v <<= 32 - w                        ' at most 32 digits
   repeat w
     out((v <-= 1) & 1 + "0")

PRI waitForCog
   repeat while byte[save+1]           'wait for idle state

PRI sendToCog(command,char)
   waitForCog                          'wait for idle state
   word[save] := command<<8 + char     'set new command to do

PRI getDispCol                         'return column in range 0-39
   waitForCog
   return byte[save+2] - 1

PRI checkNewline                       'check for end of line
   if getDispCol == cols
      byte[save+2] := 1
      if ++byte[save+3] > rows         'roll up and blank fill
         long[save] := constant(1<<24+1<<16+(moveRowUp+rows)<<8)
         waitForCog                    'wait for roll up
         long[save] := constant(rows<<24+1<<16) 'set correct row

PRI parm(n,d)                       ' Return parameter value or default
   result := d
   if vt100Cnt > n
      if vt100Def[n] == 0
         result := vt100Val[n]

PUB out(c) | n, xx, yy                 ' Display a character
' This routine implements a subset of ECMA-48 display control codes.
' Bsp                       - Backspace (destructive)
' Tab                       - Tab (every 8 columns)
' Lf                        - Line feed
' Cr                        - Carriage return
' Esc E                     - New line (like Cr/Lf)
' Esc D                     - Index (like Lf)
' Esc M                     - Reverse index (reverse Lf)
' Esc [ <params> <code>     - CSI (see below)
'                           -  <params> are of the form: 0;1;2;3
'                           -  1 to 4 semicolon delimited digit strings
' CSI @                     - Insert # spaces.  Default 1
' CSI A                     - Cursor up # lines.  Default 1
' CSI B                     - Cursor down # lines.  Default 1
' CSI C                     - Cursor forward (right) # cols.  Default 1
' CSI D                     - Cursor backward (left) # cols.  Default 1
' CSI H                     - Move cursor to <x,y>.  Defaults <1,1>
' CSI J                     - Erase lines (default is option 0)
'                           -  0 - from current line to bottom of screen
'                           -  1 - from top of screen to current line
'                           -  2 - entire screen.  Also homes cursor
' CSI K                     - Erase characters (default is option 0)
'                           -  0 - from cursor to end of line
'                           -  1 - from beginning of line to cursor
'                           -  2 - entire line containing cursor
' CSI L                     - Insert # lines at cursor.  Default 1
' CSI M                     - Delete # lines at cursor.  Default 1
' CSI P                     - Delete # characters at cursor.  Default 1
' CSI v                     - Roll up screen region <col,row,#cols,#rows>
' CSI w                     - Roll down screen region <col,row,#cols,#rows>
' CSI x                     - Clear screen region <col,row,#cols,#rows>
   case state
      0: case c & $7F                  ' State: Home
            Bsp:                       ' Backspace
               if getDispCol           'already at column 1?
                  byte[save+2]--       'if not, back up,
                  sendToCog(putDispChar+1," ") ' write a space, then
                  waitForCog           '  back up again
                  byte[save+2]--
            Tab:
               repeat                  'display spaces until
                  sendToCog(putDispChar+1," ") ' next multiple of 8
                  checkNewline         '  (minimum 1 space)
               while getDispCol & 7
            Lf:                        ' Line Feed
               waitForCog
               if ++byte[save+3] > rows 'roll up and blank fill
                  long[save] := constant(1<<24+1<<16+(moveRowUp+rows)<<8)
                  waitForCog           'wait for roll up & set row
                  long[save] := constant(rows<<24+1<<16)
            Cr:                        ' Carriage Return
               waitForCog
               byte[save+2] := 1       'cursor to start of line
            Esc:
               state := 1              ' Escape character
            other:
               sendToCog(putDispChar+1,c)
               checkNewline
      1: case c & $7F                  ' State: After ESC found
            "[": state := 2
                 vt100Cnt := 0
                 vt100Val[vt100Cnt]~   ' Start first parameter
                 vt100Def[vt100Cnt++]~~ '  and use default value
            "E": state := 0            ' Next line
                 out(Cr)
                 out(Lf)
            "D": state := 0            ' Index
                 out(Lf)
            "M": state := 0            ' Reverse index
                 waitForCog
                 if --byte[save+3] == 0 ' Go to previous row
                    xx := byte[save+2]
                    yy := byte[save+3] #> 1
                    long[save] := constant(rows<<24+1<<16+(moveRowDown+rows)<<8)
                    waitForCog         'wait for roll down & set row
                    word[save+2] := yy<<8+xx
      2: case c & $7F                  ' State: Parameters after "["
            "0".."9":
                 vt100Val[--vt100Cnt] *= 10 ' Get explicit value
                 vt100Val[vt100Cnt] += (c & $7F) - "0"
                 vt100Def[vt100Cnt++]~
            ";": if vt100Cnt < maxVt100 ' Reuse last parameter if
                    vt100Cnt++         '  too many already
                 vt100Val[--vt100Cnt]~ ' Initialize a new parameter
                 vt100Def[vt100Cnt++]~~
            "@": state := 0            ' Insert spaces
                 waitForCog
                 xx := byte[save+2]
                 yy := byte[save+3]
                 repeat parm(0,1)
                    byte[save+2] := cols
                    byte[save+1] := deleteRight+cols-xx
                    waitForCog
                    byte[save+2] := xx
                    word[save] := constant((putDispChar+1)<<8+" ")
                    waitForCog
                    byte[save+2] := xx
            "A": state := 0            ' Cursor up
                 waitForCog
                 byte[save+3] := (byte[save+3] - parm(1,1)) #> 1
            "B": state := 0            ' Cursor down
                 waitForCog
                 byte[save+3] := (byte[save+3] + parm(1,1)) <# rows
            "C": state := 0            ' Cursor forward (right)
                 waitForCog
                 byte[save+2] := (byte[save+2] + parm(0,1)) <# cols
            "D": state := 0            ' Cursor backward (left)
                 waitForCog
                 byte[save+2] := (byte[save+2] - parm(0,1)) #> 1
            "H": state := 0            ' Move cursor to position
                 waitForCog
                 byte[save+2] := (parm(0,1) #> 1) <# cols
                 byte[save+3] := (parm(1,1) #> 1) <# rows
            "J": state := 0            ' Erase lines
                 waitForCog
                 xx := byte[save+2]
                 yy := byte[save+3]
                 case parm(0,0)
                    0:                 ' Erase lines to bottom of screen
                       long[save] := yy<<24+1<<16+(fillRow+rows-yy+1)<<8+" "
                    1:                 ' Erase lines to top of screen
                       long[save] := 1<<24 + 1<<16 + (fillRow+yy)<<8 + " "
                    2:                 ' Erase entire screen & home cursor
                       long[save] := 1<<24 + 1<<16 + (fillRow+rows)<<8 + " "
                       xx := yy := 1
                 waitForCog
                 word[save+2] := yy<<8 + xx
            "K": state := 0            ' Erase characters
                 waitForCog
                 xx := byte[save+2]
                 yy := byte[save+3]
                 case parm(0,0)
                    0:                 ' Erase to end of line
                       long[save] := yy<<24+xx<<16+(putDispChar+cols-xx+1)<<8+" "
                    1:                 ' Erase to start of line
                       long[save] := yy<<24+1<<16+(putDispChar+xx)<<8+" "
                    2:                 ' Erase current line
                       long[save] := yy<<24+1<<16+(fillRow+1)<<8+" "
                 waitForCog            ' Reposition cursor
                 byte[save+2] := xx
            "L": state := 0            ' Insert lines
                 yy := byte[save+3]
                 byte[save+2] := 1     ' Cursor to left margin
                 repeat parm(0,1)      ' Roll down area # times
                    long[save] := rows<<24 + 1<<16 + (moveRowDown+rows-yy)<<8
                    waitForCog
                    long[save] := yy<<24 + 1<<16 + (putDispChar+cols)<<8 + " "
                    waitForCog
                 word[save+2] := yy<<8 + 1 ' Set cursor
            "M": state := 0            ' Delete lines
                 waitForCog
                 yy := byte[save+3]
                 repeat parm(0,1) <# rows-yy ' Roll up area # times
                    if yy < rows
                       long[save] := yy<<24 + 1<<16 + (moveRowUp+rows-yy)<<8
                       waitForCog
                    long[save] := rows<<24 + 1<<16 + (putDispChar+cols)<<8 + " "
                    waitForCog
                 word[save+2] := yy<<8 + 1 ' Set cursor
            "P": state := 0            ' Delete characters
                 waitForCog
                 xx := byte[save+2]
                 yy := byte[save+3]
                 n := cols - xx
                 repeat parm(0,1)
                    byte[save+1] := deleteLeft+n
                    waitForCog
                    long[save] := yy<<24 + n<<16 + (putDispChar+n)<<8 + " "
                    waitForCog
                    word[save+2] := yy<<8 + xx
   
{{
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}}
